from collections import UserList
import inspect
import re

import numpy as np

# DEFAULT_FUZZY_MINSCORES = {1: 100, 3: 100, 4: 90, 5: 85, 6: 80, 8: 75}
DEFAULT_FUZZY_MINSCORES = {1: 100, 3: 100, 4: 90, 5: 85, 6: 80}


def check_arg(arg, valid_args):
    """
    Function that checks if an argument is an element of a list of valid
    arguments.

    Parameters
    ----------
    arg : str
        The string to be checked.
    valid_args : iterable
        List or other iterable with valid arguments.

    Returns
    -------
    result : str
        The first element of a list with the elements in valid_args
        that produced a match for arg.

    Raises
    ------
    IndexError
        When no match was found.

    Notes
    ------
    Uses a regular expression that checks if the argument appears
    at the start of a string (case-insensitive).
    """

    try:
        find = re.compile(rf"^{arg}", re.IGNORECASE)
        return [s for s in valid_args if find.match(s)][0]
    except (ValueError, IndexError) as e:
        raise ValueError(f"invalid argument: '{arg}' must be in {valid_args}")


def check_arg_list(arg_list, valid_args):
    """
    Function that checks if the elements of a list of arguments are
    in a list of valid arguments by calling the function check_arg.

    Parameters
    ----------
    arg_list : str or iterable
        List or other iterable with arguments to be checked. If a str type
        is passed then it is converted to a list.
    valid_args : iterable
        List or other iterable with valid arguments.

    Returns
    -------
    result : list
        List with elements from valid_args that produced a match for the
        elements in arg_list.

    Raises
    ------
    IndexError
        When no match was found for any of the elements in arg_list.
    """

    if isinstance(arg_list, str):
        arg_list = [arg_list]

    return [check_arg(a, valid_args) for a in arg_list]


def check_if_nested_list(n_list, min_elements=2):
    """
    Function that checks if all elements of a list are lists with a
    minimum number of elements.

    Parameters
    ----------
    n_list : list
        The list to be checked.
    min_elements : int
        The minimum number of elements that the list(s) within n_list
        must consist of.

    Raises
    ------
    TypeError
        When n_list or any of its elements is not a list.
    ValueError
        When a list within n_list contains less than min_elements elements.
    """
    error_msg = f"Each nested element must be a list with >={min_elements} elements."
    if isinstance(n_list, list):
        for l in n_list:
            if isinstance(l, list):
                if len(l) < min_elements:
                    raise ValueError(error_msg)
            else:
                raise TypeError(error_msg)
    else:
        raise TypeError("Expected a nested list")

def parse_name_and_units(s):
    """
    This function attempts to extract the feature name and units from 
    a string. It tries to split the string at the first space character.
    The first item before the space is the feature name (verbatim). If
    the part after the space contains any parentheses, brackets or accolades
    these are stripped off and the remaining part is considered to contain
    the units. For example, it will correctly recognize the following 
    formats: 'Ca', 'Mg mg/l', 'Na (mg/l)', 'NO3 [mg N/l]', 'SO4 {mg/l S}'.
    The units for 'Ca tot mg/l' will be incorrectly inferred because the
    string contains two spaces.

    Parameters
    ----------
    s : str
        A string from which the name and units must be extracted.

    Returns
    -------
    name : str
        The feature name extracted from 's'. For the example strings above, 
        the resulting names will be 'Ca', 'Mg', 'Na', 'NO3', 'SO4' and 'Ca'.
    units: str
        The units extracted from 's'. For the example strings above, the 
        corresponding units will be '', 'mg/l', 'mg/l', 'mg N/l', 
        'mg/l S' and 'tot mg/l'.
    """

    # Split string at the first space.
    split_str = s.split(sep=None, maxsplit=1)
    # If the string contains no space separator the
    # string is considered to be the feature name. An
    # empty string is returned for the units.
    if len(split_str) == 1:
        return split_str[0], ""
    # If the string was sucessfully split, the
    # first list element is considered to be the 
    # feature name, the second the units.
    elif len(split_str) == 2:
        # Use a regular expression to strip off any
        # parentheses, brackets or accolades from the
        # string with the units.
        return split_str[0], re.sub("[\(\)\[\]\{\}]", "", split_str[1])
    
    # For empty strings the list returned by the split
    # function will be empty. Return empty strings for
    # both name and units in that case.
    return "", ""

class StringList(UserList):
    """
    Class with convenience methods for lists of strings.
    """
    def __init__(self, initlist=None):
        """
        Class initialization method. Ensures that all items are of type str.

        Parameters
        ----------
        initlist : list
            List of items to add to the StringList.
        """
        if initlist is not None: 
            super().__init__([str(s) for s in initlist])

    def replace_strings(self, r_dict):
        """
        This method modifies the elements of the StringList by
        performing a search and replace based on the elements in
        r_dict.

        Parameters
        ----------
        r_dict : dict
            Dictionary of which the keys are the strings to search for
            and the  values are the strings to replace the search values
            with.

        Raises
        ------
        TypeError
            When r_dict is not a dictionary.
        """
        if not isinstance(r_dict, dict):
            raise TypeError("Argument 'r_dict' must be of type dict")
        try:
            for key, value in r_dict.items():
                self.data = [s.replace(key, value) for s in self.data]
        except:
            pass

    def strip(self):
        """
        This method removes all leading or trailing whitespace from
        the string items in the list.
        """
        self.data = [s.strip() for s in self.data]

    def strip_parentheses(self):
        """
        This method modifies the elements of the StringList by
        removing all characters between parentheses and the
        parentheses themselves.
        """
        # The .* in the regular expression indicates zero or more
        # repetitions (as per the *) of any character except a
        # newline (as per the .)
        self.data = [re.sub("\(.*\)", "", s) for s in self.data]

    def tidy_strings(self):
        """
        This method modifies the elements of the StringList by
        (i) turning all characters to lowercase (ii) stripping all
        non-ASCII characters and (iii) removing all characters that
        are not letters, numbers or whitespace.
        """
        self.data = [s.lower() for s in self.data]
        self.data = [s.encode("ascii", "ignore").decode("ascii") for s in self.data]
        # Regular expression to filter out any character that is not
        # a numeric lowercase or uppercase symbol (the caret inside the
        # square brackets serves as the not operator).
        self.data = [re.sub("[^0-9a-zA-Z\s]", "", s) for s in self.data]


def valid_kwargs(f, **kwargs):
    """
    This function checks if one or more function arguments are valid
    keyword arguments for callable 'f'.

    Parameters
    ----------
    f : callable function
        Function for which the keyword arguments must be checked.
    kwargs : dict
        Dictionary of keyword arguments.

    Returns
    -------
    result : dict
        Dictionary with only the valid keyword arguments passed in kwargs.
    """
    valid_kwargs = inspect.signature(f).parameters
    rv = kwargs
    for kw in rv.copy():
        if kw not in valid_kwargs:
            rv.pop(kw)

    return rv


def fuzzy_min_score(s):
    """
    This function calculates the minimum score required for a valid
    match in fuzzywuzzy's extractOne function. The minimum score depends
    on the length of 's' and is calculated based on the string lengths and
    scores in the DEFAULT_MINSCORES dictionary.

    Parameters
    ----------
    s : str
        String for which the minimum score must be determined.

    Returns
    -------
    result : float
        The minimum score for 's'.
    """
    xp = list(DEFAULT_FUZZY_MINSCORES.keys())
    fp = [v for v in DEFAULT_FUZZY_MINSCORES.values()]
    # Use the interp function from NumPy. By default this function
    # yields fp[0] for x < xp[0] and fp[-1] for x > xp[-1]
    return np.interp(len(s), xp, fp)


def _wadi_style_warning(message):
    """
    This function formats the warning messages created when
    warnings.warn is called from within the _warn method of
    WadiBaseClass. Note that this function has prescribed
    kwargs, see https://docs.python.org/3/library/warnings.html#warnings.formatwarning

    Parameters
    ----------
    message : str
        The warning message to be displayed.
    """
    return f"WaDI warning: {message}\n"



import pandas as pd
import re
from wadi.base import WadiBaseClass
from wadi.mapper import MapperDict
from wadi.utils import check_if_nested_list
from wadi.unitconverter import UnitConverter

# Suppress performance warnings for that occur during harmonize because
# DataFrame can contain lots of NaNs
import warnings

warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)

# BD_FACTORS = {"delete": 0, "halve": 0.5}
# VALID_DUPLICATE_COLS_OPTIONS = ['keep_first', 'keep_all']


class Harmonizer(WadiBaseClass):
    """
    WaDI class for transforming the data to a harmonized format.
    """

    def __init__(
        self,
    ):
        """
        Class initialization method.
        """
        # Call the ancestor's initialization method to define the 
        # generic class attributes and methods.
        super().__init__()

        self._convert_units = False
        self._target_units = "mg/l"
        self._override_units = {}
        self._drop_columns = []
        self._merge_columns = []

        # self._bd_factor = BD_FACTORS[check_arg(convert_bds, BD_FACTORS.keys())]
        # Define the regular expression that searches for values below 
        # or above the detection limit.
        self._bd_RE = rf"(^\s*[<>]\s*)"
        # Define decimal separator symbol that deviates from a dot.
        self._decimal_str = ","

        # Initialize a UnitConverter object that will be used to 
        # determine the unit conversion factor.
        self._unit_converter = UnitConverter()

    def __call__(
        self,
        convert_units=False,
        target_units="mg/l",  # str, immutable
        override_units=None,
        drop_columns=None,
        merge_columns=None,
        detection_limit_symbols="<>",  # str, immutable
        decimal_str=",",  # str, immutable
    ):
        """
        This method provides an interface for the user to set the
        attributes that determine the Harmonizer object behavior.

        Parameters
        ----------
        convert_units : bool, optional
            When True, the units will be converted to the specified
            target units. Default: False
        target_units : str, optional
            The desired target units. Must be a string in a format
            that Pint can parse. Default: mg/l
        override_units : dict, optional
            A dictionary which specifies the column names for which
            'target_units' must be ignored and replaced with the
            units specified as the values of the dictionary.
        drop_columns : list, optional
            A list of columns that should not appear in the
            harmonized DataFrame.
        merge_columns : list, optional
            A nested list of columns names that will be merged into
            a single column. Each list element can contain any number
            of column names. The first element in each of the list
            elements is the column to keep. Any NaN values that
            appear in this column will be replaced by non-NaN values
            from the subsequent columns in the list element.
        detection_limit_symbols : str, optional
            A string that contains the symbols that are used for
            measurement values beyond (below or above) the detection
            limit. Default: '<>' (both values below the detection
            limit, e.g. < 0.1, and above the detection limit, e.g.
            > 0.1 will be identified this way).
        decimal_str : str, optional
            The character used as the decimal separator when a number
            is represented as a string in the original data,
            typically when it is below the detection limit, for
            example '< 0,5'. Default: ','.

        Returns
        ----------
        result : DataFrame
            A DataFrame with the transformed data.
        """
        self._convert_units = convert_units
        self._target_units = target_units
        self._override_units = override_units or {}
        self._drop_columns = drop_columns or []
        self._merge_columns = merge_columns or []
        self._decimal_str = decimal_str
        check_if_nested_list(self._merge_columns)

        self._bd_RE = rf"(^\s*[{detection_limit_symbols}]\s*)"

    def _convert_values(self, v, conversion_factor):
        """
        Convert cell values using the unit conversion factor. This
        function is used in _execute with the Pandas 'apply' function
        and is needed because (i) values with a detection limit symbol
        cannot be converted simply by multiplying and (ii) some cell
        values are imported as strings even though they are numbers.

        Parameters
        ----------
        v : float or str
            The value to be converted.
        conversion_factor : float
            The unit conversion factor.

        Returns
        ----------
        result : float
            The converted value.
        """
        # If the cell value is not a string, it is (always?) a float
        # so the function can directly return the cell value times the
        # unit conversion factor.
        if not isinstance(v, str):
            return v * conversion_factor

        # Use the regex split function to split the string. A cell value
        # like <0.5 will be parsed into the following list ['', '<', '0.5'].
        substrings = re.split(self._bd_RE, v)
        # Check if there are three substrings
        if len(substrings) == 3:
            # Convert the third substring to a float and apply the uc_factor.
            try:
                rv = float(substrings[2].replace(self._decimal_str, ".")) * conversion_factor
            except ValueError:
                rv = substrings[2]
            # Prefix the smaller-than symbol without any leading and
            # only a single trailing space and return as a string.
            return f"{substrings[1].strip()} {rv}"
        # If the split function returned only one value it can be a number.
        # In that case try to convert the substring to a float and return
        # it after applying the conversion factor.
        elif len(substrings) == 1:
            try:
                return float(substrings[0]) * conversion_factor
            # If a value error occurs the substring was not a number and
            # the input value is simply returned.
            except ValueError:
                return v
        else:
            # If all of the above fails, simply return the input value.
            return v

    def _execute(
        self,
        infotable,
    ):
        """
        This function performs the harmonize operations, that is it
        converts the units to the target units, it deletes any 
        undesired columns and merges columns.

        Parameters
        ----------
        infotable : InfoTable
            The DataObject's InfoTable. 

        Returns
        -------
        df : DataFrame
            DataFrame with the transformed data.
        """
        self._msg("Harmonizing", header=True)

        if (self._target_units == "hgc"):
               hgc_units_dict = MapperDict._create_hgc_units_dict()

        # Initialize the DataFrame to be created. Uses the target_index
        # in the InfoTable that ensures that the sampleids for
        # stacked data are uniquely defined. For wide data the
        # target_index is None, and the DataFrame will have an
        # ordinary RangeIndex.
        df = pd.DataFrame(index=infotable.target_index)

        # Create empty dictionaries that will be filled with values
        # of alias_n and alias_u, respectively, to be able to assign
        # the right values to the DataFrame columns at the end.
        column_header_dict = {}
        units_header_dict = {}

        # Iterate over items in the InfoTable. Recall that InfoTable
        # is a nested dict and key_0 is used to indicate the level-0
        # keys and dict_1 (their corresponding values) are the
        # level-1 dictionaries. This terminology is used here as well
        # for consistency with the code in infotable.py.
        for key_0, dict_1 in infotable.items():
            # Do not process items that the user has indicated
            # should be skipped.
            if key_0 in self._drop_columns:
                self._log(f"* Dropping column: {key_0}")
                continue

            # Get the datatype ('sampleinfo' or 'feature').
            datatype = dict_1["datatype"]

            self._msg(f"* Processing {key_0} ({datatype}).")

            # The actual data are Pandas Series that are a view
            # to the DataFrame created in read_data().
            values = dict_1["values"]

            # For 'wide' data, the index is a default RangeIndex
            # but for 'stacked' data a new index must be created
            # to be able to match the feature data to the samples.
            # Stacked data are identifiable from the InfoTable's
            # dict_1 keys because, unlike wide data, it will have a
            # 'sampleids' key.
            if "sampleids" in dict_1:
                # If the user specified multiple columns as the
                # sampleid (for example a sample location + a sample
                # date), the the value for 'sampleids' will be a
                # DataFrame and the new index will be a MultiIndex...
                if isinstance(dict_1["sampleids"], pd.DataFrame):
                    values = values.set_axis(
                        pd.MultiIndex.from_frame(dict_1["sampleids"])
                    )
                # ... but if only a single column identifies the
                # sampleids, then the value for 'sampleids' will be
                # a Series.
                else:
                    values = values.set_axis(dict_1["sampleids"])

                # Even if multiple columns identify the sampleids,
                # duplicate entries can still occur in the index,
                # for example when multiple measurement methods
                # have been used for a single parameter but were
                # given the same name in the original file.
                # In that case, the duplicated values are simply
                # discarded, effectively keeping only the first
                # sample.
                if any(values.index.duplicated()):
                    values = values.loc[~values.index.duplicated()]
                    # Warn the user, this is a crude approach and
                    # the user may wish to adjust the original file.
                    self._warn(
                        f"Duplicate sampleids found for {key_0}. Keeping only first occurrence."
                    )

            # Rows with NaNs can be dropped for faster processing,
            # since all Series share a common index, they are easily
            # pieced together at the end into a new DataFrame.
            values = values.dropna()

            # The column name in the new DataFrame will be the item's
            # alias...
            alias_n = dict_1["alias_n"]
            self._msg(f" - Alias: {alias_n}")
            # ... and the units will be the item's unit alias. Note 
            # that alias_u will be replaced by Pint's pretty format
            # if the units could be successfully parsed. The value
            # here is the value from the imported data file, which
            # will be used in case Pint fails to parse the units 
            # string.
            alias_u = dict_1["alias_u"]

            # Process the data depending on the datatype
            # For features, convert units and handle values
            # below the detection limit
            if datatype == "feature":
                # Define a unit conversion factor that will be used
                # if unit parsing fails.
                uc_factor = 1.0

                # Try to parse the input unit string with _str2pint. 
                # Returns a Pint Quantity objects qs (for the units) 
                # and mw (for the molar mass).
                u_str = dict_1["u_str"]
                qs, mw_formula, msg = self._unit_converter._str2pint(alias_n, u_str)
                self._log(msg)
                
                # Infer the unit alias from the short pretty
                # format string representation of qs. qs may be None
                # if the units were not properly identified by Pint.
                if (qs is not None):
                    alias_u = f"{qs.units:~P}"

                if self._convert_units:
                    # The user may wish to override the general
                    # target units, in which case the desired
                    # units were passed in the dict override_units.
                    if (self._target_units == "hgc"):
                        target_units = hgc_units_dict.get(alias_n)
                    elif key_0 in self._override_units:
                        target_units = self._override_units[key_0]
                    else:
                        target_units = self._target_units

                    # Call the UnitConverter object's get_uc method,
                    # which returns the target units as qt and the 
                    # unit conversion factor as uc. Both are Pint
                    # Quantity objects.
                    qt, uc = self._unit_converter.get_uc(qs, target_units, mw_formula)

                    if uc is not None:
                        # Only the magnitude attribute of uc is needed.
                        uc_factor = uc.magnitude
                        # qt's units attribute becomes the unit alias. 
                        # The formatted string with the ~P format specifier
                        # returns the units in Pint's "short pretty format".
                        alias_u = f"{qt.units:~P}"

                        # Update the log file
                        self._log(
                            f" - Converting units for {alias_n} from {dict_1['unit']} to {alias_u}."
                        )
                        self._log(f" - Unit conversion factor: {uc:~P}.")
                    else:
                        # Write a message to the log file about the failed unit
                        # conversion attempt.
                        self._log(
                            f" - Could not convert from {u_str} to {target_units} for {alias_n}."
                        )

                # Convert the measurement values using _convert_values.
                values = pd.to_numeric(values, errors="ignore")
                values = values.apply(self._convert_values, conversion_factor=uc_factor)

            # Add the values Series as a column to the new DataFrame.
            df[key_0] = values

            # Replace the original name and units by their aliases.
            column_header_dict[key_0] = alias_n
            units_header_dict[key_0] = alias_u

        # Check for any columns to be merged. Iterates over the lists in
        # _merge_columns. The first list element is the column of which the
        # NaN values must be replaced by the non-NaN values in the subsequent
        # columns in the list. Merging is skipped when the units of the source
        # column is not compatible with the target column
        if len(self._merge_columns):
            self._log("* Merging columns")
        for c_list in self._merge_columns:
            if not set(c_list).issubset(units_header_dict):
                self._log(f" - Merge failed for [{', '.join(c_list)}]")
                continue
            # The target column name is the first element of the list
            target_col = c_list[0]
            # Get the units string of the target column.
            target_col_units = units_header_dict[target_col]
            # Iterate over the source columns.
            for source_col in c_list[1:]:
                # Get the source column's units string.
                source_col_units = units_header_dict[source_col]
                # Check if target and source column units are compatible.
                if target_col_units == source_col_units:
                    # Replace the NaNs in the target column with values
                    # from the source column.
                    df[target_col].fillna(df[source_col], inplace=True)
                    # Write a message to the log file
                    self._log(
                        f" - NaN values in column {target_col} will be replaced with values from column {source_col}."
                    )
                    # Delete the source column from df...
                    df.drop(source_col, axis=1, inplace=True)
                    # ... and its name and units from the dicts
                    # that will be used to create df's columns.
                    column_header_dict.pop(source_col)
                    units_header_dict.pop(source_col)
                else:
                    # Log a message about the failed merged attempt.
                    self._log(
                        f" - Merge failed: Units of column {target_col} ({target_col_units}) incompatible with column {source_col} ({source_col_units})."
                    )

        # Write the logged messages to the log file
        self.update_log_file()
        
        # Assign a nested list to df.columns. This way the column
        # headings become a MultiIndex, so that both the column names
        # and the units appear above the columns in the DataFrame
        df.columns = [
            [column_header_dict[c] for c in df.columns],
            [units_header_dict[c] for c in df.columns],
        ]

        # Return the DataFrame
        return df



""" Functions to help the building of the documentation. Especially finding files is a pain in ReadTheDocs when compared
to local building of the documentation """

from pathlib import Path
import glob
import os
def get_data_dir():
    # cur_dir = Path.cwd()

    # data_file_not_found = True
    # max_directory_levels = 3
    # directory_level = 0

    # example_data_file = "**/hgc_example.xlsx"

    # while data_file_not_found:
    #     data_file = glob.glob(example_data_file, recursive=True)
    #     if len(data_file) > 0:
    #         data_file_not_found = False
    #         data_file_folder = (cur_dir / data_file[0]).parent
    #         os.chdir(cur_dir)
    #         return data_file_folder
    #     if directory_level > max_directory_levels:
    #         os.chdir(cur_dir)
    #         return None
    #     os.chdir('..')
    #     directory_level += 1
    return Path(__file__).parent / 'data_for_documentation'



from collections import UserDict
import json
import numpy as np
import pandas as pd
from pathlib import Path
import re

import fuzzywuzzy.fuzz as fwf
import fuzzywuzzy.process as fwp

from wadi.base import WadiBaseClass
from wadi.utils import StringList, check_arg_list, fuzzy_min_score
from wadi.api_utils import (
    translate_strings,
    query_pubchem_fuzzy,
    query_pubchem_cas,
    query_pubchem_synonyms,
)
from wadi.unitconverter import UnitRegexMapper

DEFAULT_STR2REPLACE = {
    "Ä": "a",
    "ä": "a",
    "Ë": "e",
    "ë": "e",
    "Ö": "o",
    "ö": "o",
    "ï": "i",
    "Ï": "i",
    "μ": "u",
    "µ": "u",
    "%": "percentage",
    "°C": "degC",
    "°F": "degF",
}
DEFAULT_FILTERSTR = [
    "gefiltreerd",
    "na filtratie",
    "filtered",
    "filtration",
    " gef",
    "filtratie",
]

DEFAULT_STR2REMOVE = [
    "icpms",
    "icpaes",
    "gf aas",
    "icp",
    "koude damp aas",
    "koude damp",  # whitespace, string
    "berekend",
    "opdrachtgever",
    "gehalte",
    "kretl",
    "tijdens meting",
    # 'gefiltreerd', 'na filtratie', 'filtered', 'gef', 'filtratie',
    "na destructie",
    "destructie",
    "na aanzuren",
    "aanzuren",
    "bij",  # whitespace, string, whitespace
]

VALID_MATCH_METHODS = ["exact", "ascii", "regex", "fuzzy", "pubchem"]


class MapperDict(UserDict):
    """
    Instances of this class are meant to be used as dictionaries
    that translate names and units into their aliases. It derives
    from the UserDict wrapper to extend the functionality of a
    regular Python dict with some class methods to create an
    instance from json files or PubChem queries. The dictionary
    can also be saved to a json file and it contains a translate
    method to create a dictionary that translates between languages
    (currently not working due to issues with Google Translate).
    """

    @classmethod
    def from_file(cls, file_path):
        """
        This method initializes the class by reading a dictionary
        from a json file.

        Parameters
        ----------
        file_path : str
            The json file to be read.

        Returns
        ----------
        result : class instance
            An instance of the UserDict class containing the
            imported data.
        """
        with open(file_path, "r") as fp:
            return cls(json.load(fp))

    @classmethod
    def default_dict(cls, keys, values):
        """
        This method initializes the class by reading a dictionary
        from the file default_feature_map.json. The function
        arguments are the column names that should become the
        dictionary keys and values, respectively.

        Parameters
        ----------
        keys : str
            The name of the column whose values should become
            the dictionary keys.
        values : str
            The name of the column whose values should become
            the dictionary values.

        Returns
        ----------
        result : class instance
            An instance of the UserDict class containing the
            imported data.
        """
        # The json file resides in the parent directory of the
        # current module's py file. The __file__ attribute returns
        # the pathname of the current py file and .parent
        # provides its parent directory.
        filepath = Path(__file__).resolve().parent
        # Import the file into a DataFrame.
        dfj = pd.read_json(Path(filepath, "mapping_data", "default_feature_map.json"))
        # Use the DataFrame's explode function to transform any keys
        # that are a list (or list-like) into a row. The corresponding
        # value is duplicated for each list element that becomes a row.
        dfd = dfj[[keys, values]].explode(keys).dropna()
        # Convert the DataFrame to a dictionary
        rv_dict = dfd.set_index(keys)[values].to_dict()
        # Values are lists, gives problems when they become aliases so
        # keep only first list element if there are multiple.
        rv_dict = {k: str(v[0]) for k, v in rv_dict.items()}
        return cls(rv_dict)

    @classmethod
    def _create_hgc_units_dict(cls):
        """
        This method initializes the class by reading the csv
        files in the folder hgc_constants, which contain
        the feature names and their target units.

        Returns
        ----------
        result : class instance
            An instance of the UserDict class containing the
            imported data.
        """

        rv_dict = {}
        filepath = Path(__file__).resolve().parent
        for fname in ["atoms.csv", "ions.csv", "other_than_concentrations.csv"]:
            df = pd.read_csv(Path(filepath, "mapping_data", fname), comment='#')
            rv_dict = {**rv_dict, **df.set_index("feature")["unit"].to_dict()}

        return cls(rv_dict)

    @classmethod
    def pubchem_cas_dict(
        cls,
        strings,
        src_lang=None,
        max_attempts=10,
    ):
        """
        This method creates a dictionary with CAS numbers for the
        names in 'strings' using the PubChem REST API.

        Parameters
        ----------
        strings : list or list-like
            A list of the names that should be looked up.
        src_lang : str, optional
            String that specifies the original language of the names
            in 'strings. Default: None.
        max_attempts : int, optional
            The maximum number of attempts to connect to the Google
            Translate API. Default: 10.

        Returns
        ----------
        result : class instance
            An instance of the UserDict class in which the elements
            of 'strings' are the keys and the CAS numbers the values.
            An empty dictionary is returned if translation of the
            strings failed for some reason.
        """
        # Convert 'strings' into a list just in case a single string
        # is passed
        strings = list(strings)
        if src_lang is not None:
            translated_keys = translate_strings(strings, src_lang, "EN", max_attempts)
        else:
            translated_keys = strings

        # Start with an empty dict
        rv_dict = {}
        if translated_keys is not None:
            rv_dict = {
                s: query_pubchem_cas(s_en) for s, s_en in zip(strings, translated_keys)
            }

        return cls(rv_dict)

    @classmethod
    def pubchem_cid_dict(
        cls,
        strings,
        src_lang=None,
        max_attempts=10,
    ):
        """
        This method creates a dictionary with CID numbers for the
        names in 'strings' using the PubChem REST API.

        Parameters
        ----------
        strings : list or list-like
            A list of the names that should be looked up.
        src_lang : str, optional
            String that specifies the original language of the names
            in 'strings. Default: None.
        max_attempts : int, optional
            The maximum number of attempts to connect to the Google
            Translate API. Default: 10.

        Returns
        ----------
        result : class instance
            An instance of the UserDict class in which the elements
            of 'strings' are the keys and the CID numbers the values.
            An empty dictionary is returned if translation of the
            strings failed for some reason.
        """
        # Convert 'strings' into a list just in case a single string
        # is passed
        strings = list(strings)
        if src_lang is not None:
            translated_keys = translate_strings(strings, src_lang, "EN", max_attempts)
        else:
            translated_keys = strings

        # Start with an empty dict
        rv_dict = {}
        if translated_keys is not None:
            for s, s_en in zip(strings, translated_keys):
                # Add each element s to the dict
                rv_dict[s] = None
                # Look up the PubChem synonyms
                res = query_pubchem_synonyms(s_en)
                # If a result was returned
                if res is not None:
                    # Look up the key 'CID' in the first element
                    # of the dict that was returned. An IndeXError
                    # may occur if either of the keys '0' or 'CID'
                    # was not returned (in which case rv_dict[s]
                    # remains None).
                    try:
                        rv_dict[s] = res[0]["CID"]
                    except IndexError:
                        pass

        return cls(rv_dict)

    @classmethod
    def translation_dict(
        cls,
        strings,
        src_lang="NL",
        dst_lang="EN",
        max_attempts=10,
    ):
        """
        This method attempts to create a mapping dictionary with
        'strings' being the keys and their translations being the
        values.

        Parameters
        ----------
        strings : list
            List with the strings to translate.
        src_lang : str, optional
            String that specifies the language to translate from.
            Default: "NL".
        dst_lang : str, optional
            String that specifies the language to translate to.
            Default: "EN".
        max_attempts : int, optional
            The maximum number of attempts to connect to the Google
            Translate API. Default: 10.
        """

        translated_keys = translate_strings(strings, src_lang, dst_lang, max_attempts)
        if translated_keys is not None:
            rv_dict = {k: v for k, v in zip(strings, translated_keys)}
            return cls(rv_dict)

    def to_file(self, file_path):
        """
        This method saves the contents of the dictionary as a json
        file.

        Parameters
        ----------
        file_path : str
            The json file to be written.
        """
        with open(file_path, "w") as fp:
            json.dump(self.data, fp, indent=2)

    def translate_keys(
        self,
        src_lang="NL",
        dst_lang="EN",
        max_attempts=10,
    ):
        """
        This method attempts to translate the dictionary keys from
        src_lang to dst_lang using Google Translate.

        Parameters
        ----------
        src_lang : str, optional
            String that specifies the language to translate from.
            Default: "NL".
        dst_lang : str, optional
            String that specifies the language to translate to.
            Default: "EN".
        max_attempts : int, optional
            The maximum number of attempts to connect to the Google
            Translate API. Default: 10.
        """

        translated_keys = translate_strings(
            self.data.keys(), src_lang, dst_lang, max_attempts
        )
        if translated_keys is not None:
            translated_dict = {
                k: v for k, v in zip(translated_keys, self.data.values())
            }
            self.data = translated_dict

    def __str__(self):
        """
        This method provides the string representation of the
        mapping dictionary.

        Returns
        ----------
        result : str
            String that provides an overview of the keys and the values of
            the mapping dictionary.
        """
        max_lines = 10
        rv = f"This dictionary contains {len(self.data)} elements.\n"
        if len(self.data) > max_lines:
            rv += f"Only the first {max_lines} elements are shown.\n"
        rv += (
            f"This mapping dictionary contains the following names and their aliases:\n"
        )
        for i, (key, value) in enumerate(self.data.items()):
            rv += f" - {key} --> {value}\n"
            if i > max_lines:
                break

        return rv


class Mapper(WadiBaseClass):
    """
    WaDI class that implements the operations to map feature names
    and units to their aliases.
    """

    def __init__(
        self,
        key_1,
    ):
        """
        Class initialization method.

        Parameters
        ----------
        key_1 : str
            The value of key_1 is either 'name' or 'unit' depending
            on what needs to be mapped. Note that these names must
            correspond to their respective keys in the dict_1
            elements of the InfoTable, hence the attribute name 'key_1'.
        """
        # Call the ancestor's initialization method to define the
        # generic class attributes and methods.
        super().__init__()

        # Set the key that defines the mapper type
        self._key_1 = key_1

        # Select appropriate default match method depending on the
        # mapper type.
        if self._key_1 == "name":
            self.match_method = ["exact"]
        elif self._key_1 == "unit":
            self.match_method = ["regex"]

        # Define the _m_dict attribute but do not assign a mapping
        # dictionary.
        self._m_dict = None

        # For parsing units, add a UnitRegexMapper object (could
        # be used for name mapping in principle as well)
        self._regex_map = UnitRegexMapper()

        # Define some of the string manipulation attributes
        self._strip_parentheses = False
        self._replace_strings = DEFAULT_STR2REPLACE
        self._remove_strings = DEFAULT_STR2REMOVE

    def __call__(
        self,
        m_dict=None,
        match_method=None,
        regex_map=UnitRegexMapper(),
        replace_strings=None,
        remove_strings=None,
        # strip_parentheses=False,
    ):
        """
        This method provides an interface for the user to set the
        attributes that determine the Mapper object behavior.

        Parameters
        ----------
        m_dict : MapperDict or dict
            The dictionary that will map the names to their alias.
        match_method : str or list
            One or more names of the match method(s) to be used to find
            feature and unit names. Valid values include 'exact', 'ascii',
            'regex', 'fuzzy', 'pubchem'. Default is 'exact' for name
            mapping or 'regex' for unit mapping.
        regex_map : UnitRegexMapper
            UnitRegexMapper object to be used for mapping when match_method
            = 'regex'.
        replace_strings : dict
            Dictionary for searching and replacing string values in in
            the feature names or units. The keys are the search strings
            and the values the replacement strings. Default:
            DEFAULT_STR2REPLACE.
        remove_strings : list
            List of strings that need to be deleted from the feature names
            or units. Default: DEFAULT_STR2REMOVE.
        """

        if self._key_1 == "name":
            self.match_method = match_method or ["exact"]
        elif self._key_1 == "unit":
            self.match_method = match_method or ["regex"]
        self.match_method = check_arg_list(self.match_method, VALID_MATCH_METHODS)

        if isinstance(m_dict, dict):
            self._m_dict = MapperDict(m_dict)
        else:
            self._m_dict = m_dict

        # For parsing units, add a UnitRegexMapper object (could
        # be used for name mapping in principle as well)
        self._regex_map = regex_map

        # Define some of the string manipulation attributes
        self._replace_strings = replace_strings or DEFAULT_STR2REPLACE
        self._remove_strings = remove_strings or DEFAULT_STR2REMOVE
        # self._strip_parentheses = strip_parentheses

    def _df2excel(self, df):
        """
        This method creates an ExcelWriter instance that will either
        append a worksheet to an existing Excel file (for example
        when units are mapped after names) or creates a new Excel
        file when it does not yet exist (when mapping is performed
        for the first time). Any sheets in an already-existing file
        will get overwritten through the use of if_sheet_exists='replace'.

        Parameters
        ----------
        df : DataFrame
            The DataFrame to be saved to the Excel file.

        Raises
        ------

        FileNotFoundError
            When the file exists but has a size of zero bytes. This
            error is caught internally and the 'mode' attribute is
            changed from 'a' to 'w' to overwrite the existing file.
        """
        # Define the Excel file name based on the name of the log file.
        xl_fpath = Path(
            self._output_dir,
            f"mapping_results_{self._log_fname.stem}.xlsx",
        )

        try:
            # Check if any existing file with the same name has a
            # size greater than zero bytes. A file may have been
            # created but not properly written if an error occured.
            # If that was the case, an error will occur.
            if xl_fpath.stat().st_size > 0:
                # If the file size is greater than zero, initialize
                # the ExcelWriter object with mode="a" to append the
                # worksheet.
                xl_writer = pd.ExcelWriter(
                    xl_fpath,
                    mode="a",
                    engine="openpyxl",
                    if_sheet_exists="replace",
                )
            else:
                # If the file size is zero, raise the same error that
                # would be raised if the file did not exists.
                raise FileNotFoundError
        except FileNotFoundError:
            # If the file does not exist, initialize the ExcelWriter
            # object with mode="w" so that the worksheet is added to
            # an empty file.
            xl_writer = pd.ExcelWriter(
                xl_fpath,
                engine="openpyxl",
                mode="w",
            )

        # Add the DataFrame to write it to the Excel file.
        df.to_excel(
            xl_writer,
            sheet_name=f"{self._key_1.capitalize()}s",
            index=False,
        )
        # Must call close otherwise any append operation will fail.
        xl_writer.close()

    def _match_exact(self, strings, m_dict):
        """
        This method returns the values of the items in m_dict for the
        elements in 'strings' that match exactly with a key.

        Parameters
        ----------
        strings : list
            A list with strings to be matched to the keys in m_dict.
        m_dict : dict
            Dictionary to look up the elements of 'strings' in.

        Returns
        -------
        result : list
            A nested list which for each element in 'strings'
            contains a two-item list, the first element being
            the key that was matched, the second element being
            the corresponding value from m_dict.
        """
        return [[s, m_dict[s]] if s in m_dict else [None, None] for s in strings]

    def _match_regex(self, strings):
        """
        This method returns the strings returned by the RegexMapper
        which tries to match and parse the elements in 'strings'
        using a regular expression.

        Parameters
        ----------
        strings : list
            A list with strings to be parsed.

        Returns
        -------
        result : list
            A nested list which for each element in 'strings'
            contains a two-item list, the first element being
            the string produced by the RegexMapper, the second
            element being None.
        """
        regex_match = re.compile(self._regex_map.RE)
        matches = [regex_match.match(s) for s in strings]
        return [
            [self._regex_map.str(m.groupdict()), None] if m else [None, None]
            for m in matches
        ]

    def _match_fuzzy(self, strings, m_dict):
        """
        This method returns the values of the keys in m_dict for the
        elements in 'strings' that have a fuzzy score above a certain
        threshold. The score is calculated by fuzzywuzzy's extractOne
        method.

        Parameters
        ----------
        strings : list
            A list with strings to be matched to the keys in m_dict.
        m_dict : dict
            Dictionary to look up the elements of 'strings' in.

        Returns
        -------
        result : list
            A nested list which for each element in 'strings'
            contains a two-item list, the first element being
            the key that was matched (alongside with the score),
            the second element being the corresponding value from
            m_dict.
        """
        # Create a lambda function to call the extractOne function
        # for an element in 'strings'.
        fuzzy_score = lambda s: fwp.extractOne(
            s,
            list(m_dict.keys()),
            scorer=fwf.token_sort_ratio,
            score_cutoff=fuzzy_min_score(s),
        )
        # Also create a lambda function that formats the name
        # of the matched string and the fuzzy score.
        tuple2str = lambda t: f"{t[0]} (score: {t[1]}%)"
        # Call the fuzzy_score lambda function for each element in
        # 'strings'.
        scores = [fuzzy_score(s) for s in strings]
        # Return the nested list with the matched keys (including
        # the score) and their corresponding values.
        return [[tuple2str(s), m_dict.get(s[0])] if s else [None, None] for s in scores]

    def _match_pubchem(self, strings):
        """
        This method tries to look up the first compound returned by
        a call to PubChem's autocomplete API and its synonym.

        Parameters
        ----------
        strings : list
            A list with strings for which to look up the PubChem
            compound and synonym.

        Returns
        -------
        result : list
            A nested list which for each element in 'strings'
            contains a two-item list, the first element being
            the first compound returned by the PubChem autocomplete
            API, the second element being the corresponding synonym.
        """
        # Call the query_pubchem_fuzzy function for each element in
        # 'strings'. Note that the function returns a list with two
        # elements.
        return [query_pubchem_fuzzy(s) for s in strings]

    def _execute(
        self,
        columns,
        strings,
    ):
        """
        This method calls the match methods specified by the user and
        saves a summary of the results in an Excel file. The results
        are passe back to WaDI's DataObject in a dictionary.

        Parameters
        ----------
        columns : list
            A list with the names of the features (for 'stacked'
            data) or columns for 'wide' data. In both cases they
            correspond to the 'key_0' items in the InfoTable.
        strings : list
            A list with strings to be matched. These can be (feature)
            names or units.

        Returns
        -------
        result : dict
            A dictionary with the aliases (for names) or parsed
            strings (for units) for each element in 'columns'.
        """
        self._msg(f"{self._key_1.capitalize()} mapping", header=True)

        # Convert the input strings to a StringList (which has special
        # methods to modify the string elements of the list).
        try:
            strings = StringList(strings)
        except:
            TypeError("Not a valid type")

        # If the user didn't specify a mapping dictionary, create one
        # using the input strings as keys (the values will all be None).
        if self._m_dict is None:
            self._m_dict = dict.fromkeys(strings)

        # Create a DataFrame that will contain a summary of the results.
        df = pd.DataFrame({"header": columns, "name": strings})

        # Use the StringList method replace_string to replace and remove
        # strings as specified by the user.
        strings.replace_strings(self._replace_strings)
        strings.replace_strings({k: "" for k in self._remove_strings})
        # Remove any leading or trailing whitespace
        strings.strip()
        # Store the modified strings in df
        df["modified"] = strings

        # Check if 'ascii' or 'fuzzy' were passed
        if any([m in ["ascii", "fuzzy"] for m in self.match_method]):
            # Create a new StringList object and call the tidy_strings,
            # (converts all characters to lowercase and removes all
            # non-ASCII characters, as well as characters that are
            # not letters, numbers or whitespace).
            strings_t = StringList(strings)
            strings_t.tidy_strings()
            # Add the tidied strings to df
            df["tidied"] = strings_t

            # Define a new mapping dictionary for which the keys are
            # tidied versions of the keys of the mapping dictionary
            # specified by the user. Repeats the same methods that
            # were used to create strings_t, but now for the keys of
            # m_dict.
            keys_t = StringList(self._m_dict.keys())
            keys_t.replace_strings(self._replace_strings)
            keys_t.replace_strings({k: "" for k in self._remove_strings})
            keys_t.strip()
            keys_t.tidy_strings()
            m_dict_t = {
                k1: self._m_dict.get(k0) for k0, k1 in zip(self._m_dict, keys_t)
            }

        # Add columns to df that will be populated after calling the
        # match method(s).
        df["searched"] = np.nan  # Stores the item that was searched,
        df["found"] = np.nan  # the item that was found,
        df["alias"] = np.nan  # the alias of the found item,
        df["method"] = np.nan  # and the method by which the match was found.

        # Iterate over the specified match methods.
        for m in self.match_method:
            # Select only the rows for which no match was found yet
            idx = df["found"].isnull()
            if all(idx == False):
                break
            dfsub = df.loc[idx, df.columns].copy()
            # The terms to be matched depends on the method
            if m in ["exact", "regex", "pubchem"]:
                dfsub["searched"] = dfsub["modified"]
            else:  # m is ascii or fuzzy
                dfsub["searched"] = dfsub["tidied"]
            # Call the appropriate match method (a bit verbose for readability)
            if m == "exact":
                res = self._match_exact(dfsub["searched"], self._m_dict)
            elif m == "ascii":
                res = self._match_exact(dfsub["searched"], m_dict_t)
            elif m == "regex":
                res = self._match_regex(dfsub["searched"])
            elif m == "fuzzy":
                res = self._match_fuzzy(dfsub["searched"], m_dict_t)
            elif m == "pubchem":
                res = self._match_pubchem(dfsub["searched"])
            else:
                raise NotImplementedError

            # Store the matched strings in the 'found' column and
            # their aliases in the 'alias' columns. Note that the
            # regex match method does not set the unit aliases.
            dfsub[["found", "alias"]] = res
            # Only place method in the 'method' column if a match
            # was found.
            idx = ~dfsub["found"].isnull()
            dfsub.loc[idx, "method"] = m
            # Update the main df with the values in dfsub.
            df.update(dfsub)

            # Write an overview of the match method's results to the
            # log file.
            self._log(f"* Match method '{m}' yielded the following results:")
            for name, searched, found, alias in zip(
                dfsub["name"], dfsub["searched"], dfsub["found"], dfsub["alias"]
            ):
                if found is not None:
                    self._log(
                        f" - {name}: Searched {searched}, found {found}, alias {alias}."
                    )

        idx = df["alias"].isnull()
        df.loc[idx, "alias"] = df.loc[idx, "modified"]

        # Write the logged messages to the log file and the DataFrame
        # to the Excel file
        self.update_log_file()
        self._df2excel(df)

        # Create the functions return value (a dictionary with the name
        # aliases or parsed unit strings).
        rv_dict = {}
        if self._key_1 == "name":
            # Transfer the aliases found to the DataObject's
            # InfoTable by iterating over the 'alias' column. The
            # values in the 'header' column match up with the level-0
            # keys of the InfoTable.
            for index, (key_0, alias) in df[["header", "alias"]].dropna().iterrows():
                rv_dict[key_0] = {"alias_n": alias}
        elif self._key_1 == "unit":
            # Transfer the unit strings found to the DataObject's
            # InfoTable by iterating over the 'found' column. The
            # values in the 'header' column match up with the level-0
            # keys of the InfoTable.
            for index, (key_0, u_str) in df[["header", "found"]].dropna().iterrows():
                rv_dict[key_0] = {"u_str": u_str}

        return rv_dict

    def match(
        self,
        strings,
    ):

        alias_dict = self._execute(strings, strings)
        df_dict = {}
        for key, value in alias_dict.items():
            alias = list(value.values())[0]
            if key == alias:
                df_dict[key] = np.nan
            else:
                df_dict[key] = alias

        return pd.DataFrame.from_dict(
            df_dict, orient="index", columns=["alias"]
        ).reset_index()



import copy
import pandas as pd

from wadi.base import WadiBaseClass
from wadi.infotable import InfoTable
from wadi.utils import check_arg, valid_kwargs

# Valid values for the 'format' kwarg
VALID_FORMATS = ["stacked", "wide", "gef"]

# Required column headers for 'stacked' format
REQUIRED_COLUMNS_S = ["SampleId", "Features", "Values", "Units"]
DEFAULT_C_DICT = {s: s for s in REQUIRED_COLUMNS_S}

# Valid values for the 'datatype' kwarg for 'wide' format
VALID_DATATYPES = ["sampleinfo", "feature"]
# Select one of the VALID_DATATYPES as the default datatype
DEFAULT_DATATYPE = VALID_DATATYPES[1]

# Default NaN values, used if user does not specify a value for the na_values
# kwarg for read_excel or read_csv
# Copied from https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html
DEFAULT_NA_VALUES = [
    "",
    "-1.#IND",
    "1.#QNAN",
    "1.#IND",
    "-1.#QNAN",
    "#N/A",
    "N/A",
    #'NA', # Commented to avoid ambiguity with Na (Sodium)
    "#NA",
    "NULL",
    "NaN",
    "-NaN",
    "nan",
    "-nan",
]


class FileReader(WadiBaseClass):
    """
    WaDI class for importing data files.
    """

    def __call__(
        self,
        file_path,
        format="stacked",  # str, immutable
        c_dict=None,
        mask=None,
        lod_column=None,
        extract_units_from_feature_name=False,
        pd_reader="read_excel",  # str, immutable
        **kwargs,
    ):

        """
        This method provides an interface for the user to set the
        attributes that determine the FileReader object behavior.

        Parameters
        ----------
        file_path : str
            The file to be read.
        format : str, optional
            Specifies if the data in the file are in 'stacked' or 'wide'
            format. Permissible formats are defined in VALID_FORMATS.
            The 'gef' format is not implemented (yet). Default: 'stacked'
        c_dict : dict, optional
            Only used when the format is 'stacked'. This dictionary maps
            column names in the file to the compulsory column names defined
            in REQUIRED_COLUMNS_S. Default: DEFAULT_C_DICT
        mask : str, optional
            Name of the column that contains True/False labels. These sometimes
            occur in stacked data files to indicate if a reported value is 
            below or above the detection limit. If a valid column name is 
            specified, the values marked with `False` are filtered out from 
            the converted DataFrame. Only used when the format is 'stacked'.
            Default: None
        lod_column : str, optional
            Name of the column that contains information about whether the 
            reported measurement value is below or above the limit of 
            detection (LOD). If a valid column name is specified, the
            symbol is prefixed to the measurement value.
            Only used when the format is 'stacked'. Default: None
        extract_units_from_feature_name : bool
            Indicates if the feature name also contains the units. Default:
            False
        pd_reader : str, optional
            Name of the Pandas function to read the file. Must be a valid
            function name. While all functions implemented in Pandas could
            be used in principle, the design of WaDI has not been tested
            for functions other than read_excel and read_csv. Default:
            'read_excel'.
        **kwargs: dict, optional
            Dictionary with kwargs for the 'pd_reader' function. The
            kwargs can be a mix of WaDI specific keywords and valid
            keyword arguments for the 'pd_reader' function.
        """

        self._file_path = file_path
        self._pd_reader = pd_reader

        # Check if user provided a valid format specifier
        format = check_arg(format, VALID_FORMATS)
        # Raise error if the format is not yet implemented
        if format in ["gef"]:
            raise NotImplementedError(f"Format option {format} not implemented yet")
        self._format = format  # for use in read_data

        # Use c_dict to look up the names of the columns with the compulsory
        # names for stacked data.
        self._c_dict = c_dict or DEFAULT_C_DICT

        self._mask = mask
        self._lod_column = lod_column
        self._extract_units_from_feature_name = extract_units_from_feature_name

        self._kwargs = copy.deepcopy(vars()["kwargs"])  # deepcopy just to be sure

    def _execute(self):
        """
        This method imports the data from a file format readable by
        Pandas. Before calling the Pandas reader function, it checks
        the kwargs specified by the user when the class object was 
        initialized. 
        """

        # Use the defaults for na_values if the user did not specify their own
        if "na_values" not in self._kwargs:
            self._kwargs["na_values"] = DEFAULT_NA_VALUES

        # Check if the user specified the 'blocks' kwarg, which
        # means that multiple dataframes must be read and joined
        if "blocks" not in self._kwargs:
            # If blocks is not in kwargs then store the kwargs in a
            # one-element list
            blocks = [self._kwargs]
        else:
            # If blocks is in kwargs then check if it's a sequence
            # before continuing
            blocks = self._kwargs["blocks"]
            if not isinstance(blocks, (list, tuple)):
                raise ValueError("Argument 'blocks' must be a list or a tuple")

        # Loop over the blocks to perform some checks for inconsistent kwargs
        for kwargs in blocks:
            # For stacked data the units and datatype are inferred
            # from c_dict when the InfoTable is created
            if (self._format == "stacked") & ("units_row" in kwargs):
                kwargs.pop("units_row")
                self._warn(
                    "Argument 'units_row' can not be used in combination with stacked format and will be ignored."
                )
            if (self._format == "stacked") & ("datatype" in kwargs):
                kwargs.pop("datatype")
                self._warn("Argument 'datatype' is ignored when format is 'stacked'.")

        # Call _read_file to import the (blocks of) data into a single 
        # DataFrame.
        df, units, datatypes = self._read_file(self._file_path, self._pd_reader, blocks)

        if self._format == "stacked":
            # Use the values in the column with name 'mask' to
            # hide the values labelled as False from view.
            if self._mask is not None:
                df = df.loc[df[self._mask]]

            if self._lod_column is not None:
                df[self._c_dict["Values"]] = df[self._lod_column] + df[self._c_dict["Values"]].astype(str)

        # Create the InfoTable dictionary that stores views to the
        # imported data as well as additional information (units,
        # data type)
        infotable = InfoTable(
            df,
            self._format,
            self._c_dict,
            units,
            datatypes,
            self._extract_units_from_feature_name,
        )

        # Write the __str__ representation of the InfoTable to the
        # log file.
        self._log(infotable)

        # Write the log string to the log file
        self.update_log_file()

        return df, infotable

    def _read_file(
        self,
        file_path,
        pd_reader_name,
        blocks,
    ):
        """
        This method calls the specified Pandas reader function to
        perform the actual data import from file_path. It imports
        a DataFrame with the data as well as lists with the
        measurement units and the datatypes (the latter two are 
        not used when the data are in 'stacked' format).

        Parameters
        ----------
        file_path : str
            The file to be read.
        pd_reader_name : str
            Name of the Pandas function to read the file.
        blocks : list
            List with keyword arguments that specify (i) the number
            of the row with the units, (ii) the datatpe and (iii) any
            kwargs for the pd_reader function. Note that (i) and (ii)
            do not apply to 'stacked' data.

        Returns
        ----------
        df : DataFrame
            Pandas DataFrame with the imported data
        units : list
            List with the units for each column read.
        datatypes: list
            List with the datatypes for each column read.

        Raises
        ------
        ValueError
            When index_col is a kwarg in one of the blocks.

        Notes
        ----------
        The return values units and datatypes are used when the
        InfoTable is created for 'wide' format data. They are not
        used when the data format is 'stacked'.
        """

        # Inform user with message on screen that reading has started (may
        # take a long time for large files)
        self._msg(f"Reading data", header=True)
        self._log(f"* Reading file {file_path} with the following Pandas call(s):")

        # Get reference to pandas reader function and determine its valid
        # keyword arguments
        pd_reader = getattr(pd, pd_reader_name)
        if not (pd_reader.__name__ in ["read_excel", "read_csv"]):
            self._warn(
                f"WaDI has not been designed to work with reader {pd_reader}. Proceed with caution."
            )

        # Start with an empty DataFrame...
        df = pd.DataFrame()
        # ... and empty lists for units and datatype.
        units = []
        datatypes = []
        # Loop over the sets of kwargs in the block(s).
        for pd_kwargs in blocks:
            # Set values for unit_row and datatype, these may be
            # overridden if the user specified a kwarg for any
            # of them.
            units_row = -1
            datatype = DEFAULT_DATATYPE
            # Loop over the user-specified kwargs.
            for kwarg in pd_kwargs.copy():  # copy() is needed to avoid a RuntimeError
                if kwarg == "units_row":
                    units_row = pd_kwargs[kwarg]
                # Check if a valid datatype was passed and convert to
                # one of the standard formats contained in VALID_DATAYPES.
                if kwarg == "datatype":
                    datatype = check_arg(pd_kwargs[kwarg], VALID_DATATYPES)
                # Index columns are not supported to avoid duplicate.
                # index errors etc.
                if kwarg == "index_col":
                    raise ValueError("Argument 'index_col' not allowed in WADI")

            # Create a verbose message for the log file.
            kws = ", ".join(f"{k}={v}" for k, v in pd_kwargs.items())
            self._log(f" - pandas.{pd_reader_name}('{file_path}', {kws})")

            # Call the requested Pandas reader function to import the data.
            df_r = pd_reader(file_path, **valid_kwargs(pd_reader, **pd_kwargs))

            # Use the pd.concat function to join the return values from the
            # pandas reader function (i.e. the DataFrames read from the file).
            df = pd.concat([df, df_r], axis=1)

            # Read the units if the user specified a valid row number, else...
            if units_row > -1:
                units += self._read_single_row_as_list(
                    file_path, pd_reader, pd_kwargs, units_row
                )
            # ... create a list of empty strings with the same length as the
            # number of columns read
            else:
                units += [""] * df_r.shape[1]

            # Make sure that the datatype for this block is copied as many
            # times as there are columns in the DataFrame that was read
            datatypes += [datatype] * df_r.shape[1]

        return df, units, datatypes

    def _read_single_row_as_list(
        self,
        file_path,
        pd_reader,
        pd_kwargs,
        row_number,
    ):
        """
        This method calls the specified Pandas reader function to
        read a single row from file_path.

        Parameters
        ----------
        file_path : str
            The file to be read.
        pd_reader_name : str
            Name of the Pandas function to read the file.
        pd_kwargs : dict
            Keyword arguments for the pd_reader function.
        row_number : int
            The (zero-based) number of the row to read.

        Returns
        ----------
        result : list
            List with the values read.
        """
        # Assemble the appropriate kwargs to read a single row,
        # this may include the user-specified keywords sheet_name
        # and usecols.
        sr_kwargs = {} # sr is shorthand for single row
        if "sheet_name" in pd_kwargs:
            sr_kwargs["sheet_name"] = pd_kwargs["sheet_name"]
        if "usecols" in pd_kwargs:
            sr_kwargs["usecols"] = pd_kwargs["usecols"]
        sr_kwargs["header"] = None
        # Note that header=None does not seem to work in conjuction
        # with skiprows! Therefore, read the data up until row 
        # row_number + 1, the units will be in the last row of the
        # dataframe returned by the reader.
        sr_kwargs["nrows"] = row_number + 1

        # Read the data, replace any NaNs with empty strings and 
        # return the last row of the DataFrame as a list
        return pd_reader(file_path, **sr_kwargs).fillna("").values[-1].tolist()



from hgc import constants

import wadi as wd

def test_correct_units():
    """ Test whether the to-hgc method is using the correct units and names
    of quantitties """
    assert True
    # assert constants == wd.used_hgc_units()







from collections import UserDict
import pandas as pd
import re

from wadi.utils import parse_name_and_units

DICT_KEYS = ['name', 
             'unit', 
             'values',
             'sampleids', # Only for stacked 
             'datatype',
             'alias_n',
             'alias_u',
             'u_str',
            ]

class InfoTable(UserDict):
    """
    Nested dictionary that stores information about the imported data.
    Note that the keys of the level-0 dictionary are referred to as
    key_0. The level-1 dictionaries is referred to as dict_1 and 
    their keys as key_1.
    
    Each dict in the InfoTable will contain the following items:
     * name: the name of the feature
     * unit: the feature's units
     * sampleids (only for stacked data): the sample identifiers
     * values: a view to the (concentration) values in self.df
     * datatype: to indicate if the data are for a feature or sampleinfo
     * alias_n: the name alias
     * alias_u: the unit alias
     * u_str: the string parsed by str2pint in the UnitConverter
    """

    def __init__(self,
        df,
        format,
        c_dict,
        units, 
        datatypes,
        extract_units_from_feature_name,
        ):
        """
        Initialization method. Creates a nested dict with information
        about the imported data.

        Parameters
        ----------
        df : DataFrame
            The DataFrame of the DataObject class instance
        format : str
            The format of the data. Must be in VALID_FORMATS defined
            in reader.py.
        c_dict : dict
            Dictionary that maps the column names of a stacked-format
            data file to the compulsory column names in WaDI as defined
            in REQUIRED_COLUMNS_S in reader.py.
        units : list of strings
            List with the units for each feature (or sampleinfo data). This
            list is created by the '_read_file' function based on the unit_row
            kwarg passed by the user to read_data.
        datatypes : list of strings
            List with the datatype for each feature (or sampleinfo data). This
            list is created by the '_read_file' function based on the datatype
            kwarg passed by the user to read_data. Datatypes must all 
            be in VALID_DATATYPES.
        extract_units_from_feature_name : bool
            Indicates if the feature name also contains the units.
        """

        # Use the c_dict to identify the names of the columns with
        # the sampleids, the features (either sample information 
        # like location or the name of a chemical substance), the
        # units and the measurement values.
        if (format == 'stacked'):
            col_s = c_dict['SampleId']
            col_f = c_dict['Features']
            col_u = c_dict['Units']
            col_v = c_dict['Values']

            # _col_s must be a list otherwise the creation
            # of a MultiIndex DataFrame in Harmonizer.harmonize
            # will fail.
            if not isinstance(col_s, list):
                col_s = [col_s]

        # Populate the InfoTable depending on the data format
        dict_0 = {}
        if (format == 'stacked'):
            # Find the unique combinations of features + units in the DataFrame
            ufus = (df[col_f] + df[col_u].fillna('')).dropna().unique()
            for fu in ufus:
                # Select a unique feature + unit combination
                idx = (df[col_f] + df[col_u].fillna('') == fu)
                # Set the key in the infotable to the first element in the
                # feature name column (all values in the column will be equal)
                key_0 = df.loc[idx, col_f].iloc[0]
                feature_name = key_0
                units_str = df.loc[idx, col_u].iloc[0]
                if extract_units_from_feature_name == True:
                    feature_name, units_str = parse_name_and_units(feature_name)
                    
                # Define the dictionary with the required data
                dict_1 = {'name': feature_name,
                          'unit': units_str, 
                          'sampleids': df.loc[idx, col_s],
                          'values': df.loc[idx, col_v],
                          'datatype': 'feature'}
                # Add dict_1 to the info_table
                dict_0[key_0] = dict_1
        elif (format == 'wide'):
            # Iterate over the columns in the DataFrame
            for i, key_0 in enumerate(df.columns):
                # Pandas adds a dot followed by a number for duplicate columns
                # so to get the real feature name, it must be removed
                feature_name = key_0
                duplicate_nr = re.search('\.\d+$', feature_name)
                if duplicate_nr:
                    feature_name = feature_name[:-len(duplicate_nr.group())]
                units_str = units[i]
                if extract_units_from_feature_name == True:
                    feature_name, units_str = parse_name_and_units(feature_name)
                # Define the dictionary with the required data
                dict_1 = {'name': feature_name,
                          'unit': units_str, 
                          'values': df[key_0],
                          'datatype': datatypes[i]}
                # Add dict_1 to the info_table
                dict_0[key_0] = dict_1

        # Add aliases to ensure an alias is always defined, 
        # even if mapping does not result in any match.
        for key_0 in dict_0.copy(): # Iterate over copy to be sure
            dict_1 = dict_0[key_0]
            dict_1['alias_n'] = dict_1['name']
            dict_1['alias_u'] = dict_1['unit']
            dict_1['u_str'] = None
            dict_0[key_0] = dict_1   
        
        # The harmonize method of Harmonizer needs target_index for
        # the DataFrame it will create.
        if (format == 'stacked'):
            sampleids = df[col_s].copy()
            self.target_index = pd.MultiIndex.from_frame(sampleids).unique()
        elif (format == 'wide'):
            self.target_index = None

        # Call the ancestor init method to initiliaze the class with
        # the dictionary just created.
        super().__init__(dict_0)

    def __setitem__(self, key, value):
        """
        Setter method for items in the dictionary. The value passed
        is a dictionary and the method checks if its keys are all
        valid. It also adds two elements ('alias_n' and 'alias_u')
        so that a name or unit alias is always defined.

        Parameters
        ----------
        key : str
            The name of the key in the dictionary to assign 'value' to.
        value : dict
            The dictionary to assign to element 'key'.

        Raises
        ----------
        TypeError
            When the keys in 'value' are not in DICT_KEYS.

        """
        # Check if all keys in 'value' are valid.
        if not all([k in DICT_KEYS for k in value]):
            raise TypeError("Invalid dict")

        # Assign 'value' to element 'key'.
        self.data[key] = value

    def list_dict1_item(self, item):
        """
        This function iterates over all the (sub-)dictionaries in
        the InfoTable and retrieves the value of element 'item'. All
        items are collected in a list using list comprehension.

        Parameters
        ----------
        item : str
            The name of the key in the dictionary to be retrieved.

        Returns
        ----------
        result : list
            List of the values stored in element 'item' in the 
            (sub-)dictionaries of the InfoTable.
        """
        return [i[item] for i in self.data.values()]

    def list_keys(self):
        """
        This function returns the level-0 keys (key0) as a list.

        Returns
        ----------
        result : list
            List of the level-0 keys of the InfoTable.
        """
        return list(self.data.keys())

    def __str__(self):
        """
        This method provides the string representation of the
        InfoTable that is printed to the WaDI log file.

        Returns
        ----------
        result : str
            String that describes the contents of the InfoTable.
        """
        rv = ""
        for datatype in ['sampleinfo', 'feature']:
            s = ""
            rv += f"\n* The following {datatype} data were imported:\n"
            for key_0, dict_1 in self.data.items():
                if dict_1['datatype'] == datatype:
                    s += f" - name: {key_0}, unit: {dict_1['unit']}\n"
            if len(s):
                rv += s
            else:
                rv += " - None\n"
        
        return rv[:-1] # Remove trailing carriage return

    def update_items(self, a_dict):
        """
        This method iterates over the level-0 items of the dictionary
        and updates the level-1 dictionaries with the items in a_dict.

        Parameters
        ----------
        a_dict : dict
            Nested dictionary of which they keys correspond to the 
            level-0 keys of the current dictionary. Its values are
            level-1 dictionaries. The level-1 items of the current
            dictionary will be updated with those from a_dict. The
            item values of the current dictionary are not overwritten
            if an item_value in a_dict is None.
        """
        for key_0, dict_1 in a_dict.items():
            for key_1, value in dict_1.items():
                if value is not None:
                    self.data[key_0][key_1] = value
                        




import molmass as mm
from molmass.molmass import FormulaError
from pint import UnitRegistry
from pint.errors import DimensionalityError, UndefinedUnitError, OffsetUnitCalculusError

from wadi.api_utils import get_pubchem_molecular_weight

# DEFAULT_RE_DICT0 is the default dictionary to create a regular 
# expression for matching chemical concentration units. It is 
# designed to recognize variants such as mg/l, mg N/l or mg/l N. 
# Each element is a list, the first element being the character set 
# to match, the second element being the separator that follows.
DEFAULT_RE_DICT0 = {
    # Elements before the separator:
    # - descriptor (e.g. mg in mg/l) 
    "num": ["[a-zA-Z]*", "\s*"], 
    # - formula for molar mass (e.g. N in mg N/l) 
    "mw0": ["[a-zA-Z0-9]*", "?\s*"], 
    # Separator:
    "div": ["[/.,]", "\s*"], 
    # Elements following the separator:
    # - a number (e.g. 50 in cfu/50ml)
    "den0": ["[0-9]*", "?"], 
    # - descriptor (e.g. l in mg/l)
    "den1": ["[a-zA-Z]*", "\s*"], 
    # - formula for molar mass (e.g. NO3 in mg/l NO3)
    "mw1": ["[a-zA-Z0-9]*", "?"], 
}
# DEFAULT_RE_DICT1 is a dictionary to create a regular 
# expression that matches simple text.
DEFAULT_RE_DICT1 = {"txt": ["[a-zA-Z]*", ""]}


def dict2str(groupdict):
    """
    This function attempts to create a string that can be parsed by
    Pint based on the elements of a dictionary that is returned by
    the regex match method in Mapper._match_regex. This function works
    only for the default dictionaries defined above and creates a
    string specifically formatted for the _str2pint method in the 
    UnitConverter class.
    """

    # Store the elements of the dictionary as local variables for 
    # shorter code lines
    n = groupdict["num"]
    d0 = groupdict["den0"]
    d1 = groupdict["den1"]
    txt = groupdict["txt"]
    w0 = groupdict["mw0"]
    w1 = groupdict["mw1"]

    # Set return value to None if the groupdict elements do not
    # contain any unit information or if the 'txt' element has
    # length zero (happens when an empty string was matched).
    rv = None

    # Check if at least a numerator and/or one of the denominators
    # is contained in the dictionary with matched terms. 
    if not all(x is None for x in [n, d0, d1]):
        # Create an empty formatted string
        rv = f""
        # Add the numerator to rv if a match was found for it, else add '1'.
        if len(n):
            rv += f"{n}"
        else:
            rv += f"1"
        # Add a '/(' and the first denominator to rv if a match was found for
        # it, else add '/ (1'.
        if len(d0):
            rv += f" / ({d0}"
        else:
            rv += f" / (1"
        # Add the second denominator and a closing parenthesis to rv 
        # if a match was found for it, else add just a closing parenthesis.
        if len(d1):
            rv += f"{d1})"
        else:
            rv += f")"
        # Add the formula that was found for the molar mass (if any was matched). 
        # It is separated from the unit string by a vertical bar, which is what
        # the _str2pint method in the UnitConverter class expects.  
        if len(w0):
            rv += f"|{w0}"
        elif len(w1):
            rv += f"|{w1}"
    elif txt is not None: # Return any text string if one was matched.
        if len(txt) > 0:
            rv = txt

    return rv


class UnitRegexMapper:
    """
    Class to make working with regular expressions to match units a little
    bit easier (not easy).
    """

    def __init__(self, *args, func=dict2str):
        """
        Class initialization method. Defines the dictionaries used to create
        the regular expressions to match the unit strings with, as well as the 
        function to translate the groupdicts returned by the match method to a
        string that can be used by the _str2pint method in the UnitConverter 
        class.
        """
        # Check if any dictionaries were passed as arguments, else use the 
        # default dictionaries defined above.
        if args:
            self.RE = self._dict2RE(*args)
        else:
            self.RE = self._dict2RE(
                DEFAULT_RE_DICT0,
                DEFAULT_RE_DICT1,
            )
        # Set the function to be used for translating the groupdicts returned
        # by the match method to a string for the _str2pint method in the 
        # UnitConverter class.
        self.func = func

    @staticmethod
    def _dict2RE(*args):
        """
        This function creates a string with a regular expression for parsing
        (chemical concentration) units. For more details on group names see 
        https://docs.python.org/3/library/re.html
        """
        rv = r""
        for i, re_dict in enumerate(args):
            if not isinstance(re_dict, dict):
                raise TypeError(f"argument {re_dict} must be of type dict")
            if i == 0:
                rv += r"^\s*"
            else:
                rv += r"|^\s*"
            for key, value in re_dict.items():
                rv += rf"(?P<{key}>{value[0]}){value[1]}"
            rv += r"\s*$"

        return rv

    def str(self, groupdict):
        """
        This function is simply a wrapper that returns the string
        created by the function that translates the groupdicts 
        returned by the match method to a string that can be used by
        the _str2pint method in the UnitConverter class."""
        return self.func(groupdict)


class UnitConverter:
    """
    Class with some methods for unit parsing and conversion with Pint.
    """
    def __init__(self):
        """
        Class initialization method.
        """

        super().__init__()

        self._ureg = UnitRegistry()
        self._ureg.default_format = "~"

    def _get_mw(
        self,
        s,
    ):
        """
        This method uses the molmass library to determine the
        molar mass of a substance.

        Parameters
        ----------
        s : str
            Name of the substance.

        Returns
        ----------
        result : Pint Quantity object
            The molar mass in g/mole, or None if  a FormulaError
            was raised from within the molmass library.
        """
        rv = None
        try:
            rv = mm.Formula(s).mass * self._ureg("g/mol")
        except FormulaError as e:
            print(f" - Could not retrieve molar mass {s} with molmass library. Trying PubChem...")
            rv = get_pubchem_molecular_weight(s) 
            if rv is not None:
                rv = rv * self._ureg("g/mol")
            
        return rv

    def get_uc(
        self,
        qs,
        target_units,
        mw_formula,
    ):
        """
        Use Pint to determine the value of the unit
        conversion factor.

        Parameters
        ----------
        qs : Pint Quantity object
            The source units (for example 1 mg/l).
        target_units : str 
            String that defines the target units.
        mw_formula : str
            Chemical formula with which to convert
            between mass and molar concentration units.

        Returns
        -------
        qt : Pint Quantity object
            The target units
        uc : Pint Quantity object
            The unit conversion factor.
        """
        try:
            # Convert the target_units string to a Pint Quantity object
            qt = self._ureg(target_units)

            # Determine the molecular mass in g/mol.
            mw = None
            # Only try to determine the molecular mass if the target units
            # are molar units, which can be identified using the Pint
            # Quantity's object dimensionality property. If this is the case
            # a key called '[substance]' will be present.
            if ('[substance]' in qt.dimensionality.keys()) and (mw_formula is not None):
                mw = self._get_mw(mw_formula)
                if mw is None:
                    mw = get_pubchem_molecular_weight(mw_formula)

            # Use the source units 'to' method to determine the unit
            # conversion factor.
            if mw is None:
                uc = qs.to(qt)
            else:
                uc = qs.to(qt, "chemistry", mw=mw)

            # Divide uc by qs to get the right dimensions
            uc /= qs
        except (AttributeError, DimensionalityError, OffsetUnitCalculusError) as e:
            qt = None
            uc = None

        return qt, uc

    def _str2pint(
        self,
        name,
        u_str,
    ):
        """
        This function parses the three-part string that is created by
        _match_regex when the units are mapped using a regular
        expression.

        Parameters
        ----------
        name : str
            The feature name alias. Also serves as an alternative string
            to determine the molar mass if Pint fails to parse u_str.
        u_str : str
            String representation of the units to be parsed.

        Returns
        ----------
        uq : Pint Quantity object
            The units represented as Pint Quantity object.
        mw_formula : str
            The chemical formula of the substance, to be used in get_uc
            to convert between mass and molar concentration units.
        msg : str
            A message intended for the log file.

        Notes
        ----------
        The function uses the partition function to split the
        string at the | symbol into a three-part tuple that
        contains (i) the part before the separator, (ii) the
        separator itself (redundant, not used), and (iii) the
        part after the separator.
        The part after the | symbol may or may not contain the
        formula for the molecular mass, depending on the format
        for the units in the input file. In that case name of
        the feature (passed to the function as 'name') is
        used as the formula to determine the molecular mass.
        Both the units and the molecular mass are converted to a
        Pint Quantity object (i.e., the product of a unit and
        a magnitude).
        """
        try:
            # Use the partition to split the string at the | symbol.
            s_parts = u_str.partition("|")
            # Convert the units string to a Pint Quantity object
            uq = self._ureg.Quantity(s_parts[0])

            # Store the substance formula in mw_formula.
            mw_formula = s_parts[2]
            # If no formula was specified the length of mw
            # will be zero and in that case 'name' will be
            # used instead to look up the molecular mass.
            if len(mw_formula) == 0:
                mw_formula = name

            # Write a message to the log file
            msg = f" - Successfully parsed unit '{u_str}' with pint for {name}"

            return uq, mw_formula, msg
        except (AttributeError, TypeError, UndefinedUnitError, ValueError):
            # When an error occurs, write a message to the log file and
            # return empty return values.
            msg = f" - Failed to parse unit '{u_str}' with Pint for {name}"
            return None, None, msg



import glob
import os
from setuptools import setup, find_packages

def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()

setup(
    name='wadi',
    version='',
    packages=find_packages(exclude=['tests*']),
    data_files=glob.glob('mapping_data/**'),
    license='MIT',
    description='Generic importer for water quality data of the (Dutch) water laboratory',
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: Hydrology',
    ],
    python_requires='>=3.8',
    project_urls={
    'Source': 'https://github.com/KWR-Water/wadi',
    'Documentation': 'http://wadi.readthedocs.io/en/latest/',
    'Tracker': 'https://github.com/KWR-Water/wadi/issues',
    'Help': 'https://github.com/KWR-Water/wadi/issues'
    },
    install_requires=[
        'pandas>=1,<2',
        'pint',
        'requests',
        'molmass',
        'openpyxl>=3.0.0',
        'googletrans==3.1.0a0',
        'fuzzywuzzy',
        'levenshtein>=0.20.9',
        'python-Levenshtein>=0.20.9',
        'requests'
        ],
    include_package_data=True,
    url='https://github.com/KWR-Water/wadi',
    author='KWR Water Research Institute',
    author_email='martin.korevaar@kwrwater.nl, martin.van.der.schans@kwrwater.nl',
)



from wadi.base import WadiBaseClass
from wadi.harmonizer import Harmonizer
from wadi.mapper import Mapper
from wadi.filereader import FileReader

DEFAULT_OUTPUT_DIR = "wadi_output"

class DataObject(WadiBaseClass):
    """
    Class for importing hydrochemical data in a variety of formats. The
    class provides various functions that allow the user to convert the
    data into different formats.

    Examples
    --------

    import wadi as wd

    wdo = wd.DataObject()
    wdo.file_reader('chem_data.xlsx')
    df = wd.get_frame()
    """

    def __init__(
        self,
        log_fname="wadi.log",
        output_dir=DEFAULT_OUTPUT_DIR,
        silent=False,
    ):
        """
        Class initialization method. Initializes the parent class
        object so that a log file and an output directory are
        created. Also initializes the Reader, Mapper and Harmonizer
        callable classes so that the user can set their attributes
        via the methods file_reader, name_map, unit_map and
        harmonizer.

        Parameters
        ----------
        log_fname : str, optional
            Name of the log file. Default: 'wadi.log'
        output_dir : str, optional
            Name of the directory with output files. Default:
            'wadi_output'
        silent : bool, optional
            Flag to indicate if screen output is desired during
            data processing. When True then no screen output is
            displayed. Default is False (recommended for large
            data files when processing can be slow). When
            True messages will still appear in the log file. Warnings
            are always displayed on the screen regardless of the
            value for 'silent'.
        """

        # Call the ancestor's initialization method to set the log 
        # file name and output directory. The argument create_file is
        # set to True to ensure that the log_file is written so that
        # other parts of the code can append to it.
        super().__init__(
            log_fname,
            output_dir,
            silent,
            create_file=True,
        )

        # Define placeholder attribute for the DataFrame that will
        # contain the imported data...
        self._imported_df = None
        # # ... as well as the converted data
        self._converted_df = None

        # Define placeholder attribute for the InfoTable. The InfoTable
        # is a dict with information about column names, units, datatypes
        # and values. It is initialized when the _execute method of the
        # FileReader instance is called and used by the Harmonizer
        # to produce the converted DataFrame
        self._infotable = None

        # Initialize the FileReader object. The FileReader class
        # is designed to be callable so the self.file_reader attribute
        # becomes a method that the user can call.
        self.file_reader = FileReader()

        # Initialize Mapper objects. The Mapper class is designed
        # to be callable so the self.name_map and self.unit_map
        # attributes becomes methods that the user can call. The
        # 'name' and 'unit' arguments correspond to the keys in the
        # InfoTable class instance.
        self.name_map = Mapper("name")
        self.unit_map = Mapper("unit")

        # Initialize Harmonizer object. The Harmonizer class is
        # designed to be callable so the self.harmonizer
        # attribute becomes a method that the user can call.
        self.harmonizer = Harmonizer()

        # Loop over all attributes and check which are a subclass of
        # WadiBaseClass. If they are, ensure that the log file name
        # output directory and silent flag are all set to the same
        # value as the current object.
        for v in self.__dict__.values():
            if issubclass(type(v), WadiBaseClass):
                v._log_fname = self._log_fname
                v._output_dir = self._output_dir
                v._silent = self._silent

    def _execute(self, import_only=False):
        """
        This method calls the _execute methods of the child objects
        that read, map and harmonize the data. Upon success the
        converted data are stored in self._converted_df

        Parameters
        ----------
        import_only : bool, optional
            When True the data are read but not mapped or harmonized.
            Default: False.
        """
        # Import data
        self._imported_df, self._infotable = self.file_reader._execute()

        # Exit the function when only the imported data were requested.
        # Note that this function is not supposed to return any values.
        if import_only == True:
            return

        # Map names
        info_table_items = self.name_map._execute(
            self._infotable.list_keys(), self._infotable.list_dict1_item("name")
        )
        self._infotable.update_items(info_table_items)

        # Map units
        info_table_items = self.unit_map._execute(
            self._infotable.list_keys(), self._infotable.list_dict1_item("unit")
        )
        self._infotable.update_items(info_table_items)

        # Harmonize
        self._converted_df = self.harmonizer._execute(self._infotable)

    def get_converted_dataframe(self,
            include_units=True,
            force_conversion=False,
        ):
        """
        This method converts the input data based on the specified
        name and unit maps and harmonizer, and returns the result
        as a DataFrame.

        Parameters
        ----------
        include_units : bool, optional
            When True the DataFrame's columns will be a MultiIndex
            that contains both the feature aliases and their units.
            When set to False a DataFrame is returned of which the 
            columns simply correspond to the feature aliases and 
            the units are discarded. The latter option is useful when
            the DataFrame is intended for further processing in HGC.
            Default: True.
        force_conversion: bool, optional
            When True, the function will always map and harmonize
            the data before it returns the DataFrame. When False,
            the results from any previously executed data mapping
            and harmonizing are returned, when present. Default:
            False. 

        Returns
        ----------
        result : DataFrame
            The converted DataFrame.
        """
        if (self._converted_df is None) or (force_conversion == True):
            self._execute()

        if include_units == True:
            return self._converted_df
        else:
            level0_cols = self._converted_df.columns.get_level_values(0)
            return self._converted_df.set_axis(level0_cols, axis='columns')
    
    def get_imported_dataframe(self):
        """
        This method returns the imported DataFrame (that is, the data 
        'as read').

        Returns
        ----------
        result : DataFrame
            The imported DataFrame.
        """
        if self._imported_df is None:
            self._execute(import_only=True)
        return self._imported_df

    def get_imported_names(self):
        """
        This method returns the names of the features
        in the imported DataFrame.

        Returns
        ----------
        result : list
            A list with feature names in the imported DataFrame.
        """
        if self._infotable is None:
            self._execute(import_only=True)
        return self._infotable.list_dict1_item('name')



from abc import ABC, abstractmethod
import codecs
from datetime import datetime as dt
import os
from pathlib import Path
import warnings

from wadi.utils import _wadi_style_warning

class WadiBaseClass(ABC):
    """
    Base class for WADI DataObject class and its children.
    Defines functions to provide functionality to log and
    print messages and warnings.
    """

    def __init__(
        self,
        log_fname='wadi.log',
        output_dir='wadi_output',
        silent=False,
        create_file=False,
    ):
        """
        Class initialization method.

        Parameters
        ----------
        log_fname : str, optional
            Name of the log file
        output_dir : str, optional
            Name of the directory with output files
        silent : bool, optional,
            Flag to indicate if screen output is desired during
            data processing. When True then no screen output is
            displayed. Default is False (recommended for large 
            data files when processing can be slow). When
            True messages will still appear in the log file. Warnings
            are always displayed on the screen regardless of the
            value for 'silent'. 
        create_file : bool, optional
            Flag to indicate if the log file must be created. Must be
            explicitly set to True when an object of this class is
            initialized for the first time. In WaDI this happens in
            the init method of DataObject.
        """

        # Convert output_dir to a Path object and get the absolute
        # path by calling resolve
        self._output_dir = Path(output_dir).resolve()
        # Convert log_fname to a Path object
        self._log_fname = Path(log_fname)
        # Check if only a filename or a full path was specified for
        # the log filename. If only a filename was specified then
        # the parts argument will be equal to one. In that case the
        # output_dir is prefixed to the log filename
        if len(self._log_fname.parts) == 1:
            self._log_fname = Path(self._output_dir, self._log_fname)

        # Create subdirectory for output (log files, Excel files)
        self._output_dir.mkdir(exist_ok=True)

        self._silent = silent

        # Add the _log_str attribute. Descendants of this class will
        # use this attribute to create the text strings that will be
        # written to the log file
        self._log_str = ""
        # The first time the class is initialized the create_file
        # argument must be set to True to write the file. A single
        # line with a timestamp will be written. Note that _log_str
        # is reset to "" by the function update_log_file
        if create_file:
            self._log("WaDI log file", timestamp=True)
            self.update_log_file(mode="w")
    
    @abstractmethod
    def _execute(self):
        """
        This method must be implemented by all WaDI classes that derive
        from this class.
        """
        pass

    def _log(
        self,
        s,
        timestamp=False,
        header=False,
    ):
        """
        Appends a new line to the log file string.

        Parameters
        ----------
            s : str
                Text string to be written to log file.
            timestamp: bool
                If True the text string is followed by the current time
                between parentheses. No timestamp is appended when False.
            header: bool
                If True leading and trailing carriage returns and a
                banner will be added to the text string.
        """
        if header:
            self._log_str += f"\n{'=' * len(s)}\n"

        if timestamp:
            ts = dt.now().strftime("%d/%m/%Y %H:%M:%S")
            self._log_str += f"{s} ({ts})\n"
        else:
            self._log_str += f"{s}\n"
        
        if header:
            self._log_str += f"{'=' * len(s)}\n\n"

    def _msg(
        self,
        s,
        **kwargs,
    ):
        """
        Prints a message to the screen and appends it as a new line to the
        log file string.

        Parameters
        ----------
            s : str
                Text string to be printed to the screen and written to
                the log file.
            **kwargs : dict, optional
                Any keyword arguments to be passed onto _log_str.
        """

        if not self._silent:
            print(s)
        self._log(s, **kwargs)

    def _warn(
        self,
        s,
    ):
        """
        Throws a UserWarning and appends it as a new line to the
        log file string.

        Parameters
        ----------
            s : str
                Text string to be passed to the UserWarning and written to
                the log file.
        """

        # Override the standard formatting of warnings on the screen.
        #warnings.formatwarning = _wadi_style_warning
        warnings.warn(s) 
        self._log_str += f"Warning: {s}\n"

    def update_log_file(
        self,
        mode="a",
    ):
        """
        Saves the log file string to a text file

        Parameters
        ----------
            mode : str
                Python file mode for opening the file. Default is 'a',
                which means that the log file string will be appended
                to the log file.
        """

        try:
            # Uses the file open method from codecs to enable unicode
            # characters to be written to the file.
            with codecs.open(self._log_fname, mode, encoding="utf8") as f:
                f.write(self._log_str)
        except FileNotFoundError:
            self._warn(f"Log file {self._log_fname} could not be created.")

        # Reset _log_str to an empty string
        self._log_str = ""

    def _remove_log_file(self):
        """
        Attempts to delete the log file
        """
        try:
            os.remove(self._log_fname)
        except OSError:
            pass



import re
import requests
import time

import googletrans as gt

"""
This file contains several functions to interact with the PubChem and 
Google Translate APIs. Interaction with the PubChem REST API is directly
through the use of the requests module. The URL paths are based on the
description of the API found # here (Last consulted on 15 January 2023): 
https://pubchem.ncbi.nlm.nih.gov/docs/pug-rest.

For Google Translate, the googletrans package is used. Note that version 
3.0.0 of this package no longer seems to work, so version 3.1.0 or higher
is needed.
"""

# Define some global constants for the PubChem methods
RE_CAS = r"^(CAS-)?(\d+)-(\d+)-(\d+)$"  # Regular expression for CAS number in PubChem list of synonyms
REQ_PER_SEC = 5  # PUBCHEM API does not accept >5 requests per second
# List of property names available in PubChem when querying compound properties. See
# https://pubchem.ncbi.nlm.nih.gov/docs/pug-rest#section=Compound-Property-Tables 
PUBCHEM_COMPOUND_PROPS = [  
        "MolecularFormula",
        "MolecularWeight",
        "CanonicalSMILES",
        "IsomericSMILES",
        "InChI",
        "InChIKey",
        "IUPACName",
        "Title",
        "XLogP",
        "ExactMass",
        "MonoisotopicMass",
        "TPSA",
        "Complexity",
        "Charge",
        "HBondDonorCount",
        "HBondAcceptorCount",
        "RotatableBondCount",
        "HeavyAtomCount",
        "IsotopeAtomCount",
        "AtomStereoCount",
        "DefinedAtomStereoCount",
        "UndefinedAtomStereoCount",
        "BondStereoCount",
        "DefinedBondStereoCount",
        "UndefinedBondStereoCount",
        "CovalentUnitCount",
        "Volume3D",
        "XStericQuadrupole3D",
        "YStericQuadrupole3D",
        "ZStericQuadrupole3D",
        "FeatureCount3D",
        "FeatureAcceptorCount3D",
        "FeatureDonorCount3D",
        "FeatureAnionCount3D",
        "FeatureCationCount3D",
        "FeatureRingCount3D",
        "FeatureHydrophobeCount3D",
        "ConformerModelRMSD3D",
        "EffectiveRotorCount3D",
        "ConformerCount3D",
        "Fingerprint2D",
    ]


def translate_strings(
    strings,
    src_lang,
    dst_lang,
    max_attempts,
):
    """
    This method attempts to translate a list of strings from
    src_lang to dst_lang using Google Translate.

    Parameters
    ----------
    src_lang : str
        String that specifies the language to translate from.
        Default: "NL".
    dst_lang : str
        String that specifies the language to translate to.
        Default: "EN".
    max_attempts : int
        The maximum number of attempts to connect to the Google
        Translate API. Default: 10.
    """

    strings = list(strings)

    if not all([l.lower() in gt.LANGUAGES for l in [src_lang, dst_lang]]):
        raise ValueError("Invalid language(s) specified")

    t = gt.Translator()
    for i in range(max_attempts):
        try:
            rv = []
            for s in strings:
                rv.append(t.translate(s, src=src_lang, dest=dst_lang).text)
            return rv
        except:
            print(
                f"Failed attempt ({i}) to connect to Google Translate API. Retrying..."
            )

    return None


def get_pubchem_json(url):
    """
    This method interacts with the PubChem REST API by sending
    a URL and, when successful, returns a json dict.

    Parameters
    ----------
    url : str
        The URL containing the query/request.

    Returns
    -------
    result : dict or None
        Returns the json dictonary returned by the PubChem API, or
        None if an error occurred.
    """
    try:
        # Send the request and try to get the json dictionary
        # from the response
        response = requests.get(url, timeout=(2, 5))
        rv = response.json()
        # Any fault on the PubChem side will result in a
        # key 'Fault'. If this key exists, return None.
        if "Fault" in rv:
            return None
        # Call raise_for_status, which raises an exception
        # when the request was unsuccessful.
        response.raise_for_status()
    except Exception as e:
        # Handle the exception raised by raise_for_status(). Print a
        # message to the screen and return None.
        print("An error occured during contacting of the PubChem Power User Gateway.")
        return None
    else:
        # Pause to avoid exceeding the permitted number of requests
        # per second.
        time.sleep(1 / REQ_PER_SEC)
        # Return the json dictionary.
        return rv


def query_pubchem_fuzzy(s):
    """
    This method uses the PubChem REST auto-complete API service
    that supports fuzzy matching of input strings.

    Parameters
    ----------
    s : str
        The string to look up

    Returns
    -------
    result : list
        A two-element list. The first element is the first compound name
        returned if the request was successful, the second the first
        synonym returned by calling query_pubchem_synonyms. Returns
        [None, None] if any error occurred.
    """
    # Insert 's' into the URL for the auto-complete service.
    url = (
        f"https://pubchem.ncbi.nlm.nih.gov/rest/autocomplete/compound/{s}/json?limit=3"
    )
    # Get the json dict, return [None, None] if an error occurred.
    js = get_pubchem_json(url)
    if js is None:
        return [None, None]
    # Check the value of the 'total' key, which will be greater than zero
    # if the fuzzy search yielded any compounds.
    if js["total"] > 0:
        # Grab the first compound from the list of compounds in the dict.
        compound = js["dictionary_terms"]["compound"][0]
        # Try to determine its synonym.
        synonym = query_pubchem_synonyms(compound)
        # The return value is a list, the synonyms are in the dict that
        # is the first element of the list. Only the first synonym from
        # the list is used (there may be many).
        if isinstance(synonym, list):
            synonym = synonym[0].get("Synonym")[0]
            # Return the fuzzy-matched compound and its synonym.
            return [compound, synonym]

    # If this part of the code gets reached then the compoud/synonym
    # pair could not be returned, so return [None, None]
    return [None, None]


def query_pubchem_synonyms(s, namespace="name"):
    """
    This method uses the PubChem REST API service to look up synonyms
    of a PubChem compound based on its name or other allowed namespace.
    See https://pubchem.ncbi.nlm.nih.gov/docs/pug-rest#section=Input.

    Parameters
    ----------
    s : str
        The string to look up
    namespace : str, optional
        The namespace to seach for (for example, 'name', 'cid',
        'smiles', ...). Default: 'name'

    Returns
    -------
    result : list
        The return value is the list that is stored in the 'Information'
        item of the json dict. This list contains a dict (or perhaps more?)
        with the CID and a list of synonyms. Returns None if an error
        occurred.
    """
    # Insert 's' and 'namespace' into the URL.
    url = f"https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/{namespace}/{s}/synonyms/json"
    # Get the json dict, return None if an error occurred.
    js = get_pubchem_json(url)
    if js is None:
        return None
    # Return the list that contains the dict(s?) with CID and synonyms list.
    return js["InformationList"]["Information"]


def query_pubchem_cas(s, namespace="name", allow_multiple=False):
    """
    This method uses the PubChem REST API service to look up the CAS
    of a PubChem compound based on its name or other allowed namespace.
    See https://pubchem.ncbi.nlm.nih.gov/docs/pug-rest#section=Input.
    The CAS number returned is the first number matched by a regular
    expression in the list of synonyms for the compound.

    Parameters
    ----------
    s : str
        The string to look up
    namespace : str, optional
        The namespace to seach for (for example, 'name', 'cid',
        'smiles', ...). Default: 'name'
    allow_multiple : bool, optional
        When False only the first CAS that was identified is returned.
        When True all CAS numbers in the synonym list are returned. Default:
        False.

    Returns
    -------
    result : list
        The return value is the list that is stored in the 'Information'
        item of the json dict. This list contains a dict (or perhaps more?)
        with the CID and a list of synonyms. Returns None if an error
        occurred.
    """
    # Get the synonyms for compound s
    js_info = query_pubchem_synonyms(s, namespace)
    # If a list was returned then get the list with synonyms from the
    # dict that is the first list element.
    if (js_info is not None) and (len(js_info)):
        synonyms = js_info[0]["Synonym"]
        # If the list contains any synonyms...
        if len(synonyms) > 0:
            # ... collect the elements that are identified as a CAS number
            # by the RE_CAS regular expression.
            find = re.compile(RE_CAS, re.IGNORECASE)
            rv = [find.match(syn).groups()[-3:] for syn in synonyms if find.match(syn)]
            # Convert the three-element lists within rv to their string representation.
            rv = ["-".join(rvi) for rvi in rv]
            # If only a single CAS number is allowed return the first item, else
            # return the entire list.
            if allow_multiple:
                return rv
            else:
                return rv[0]
    else:
        # If an error occured return None
        return None


def get_pubchem_molecular_weight(s, namespace="name"):
    """
    This method uses the PubChem REST API service to look up the molecular
    weight of a PubChem compound based on its name or other allowed namespace.
    See https://pubchem.ncbi.nlm.nih.gov/docs/pug-rest#section=Input.

    Parameters
    ----------
    s : str
        The string to look up
    namespace : str, optional
        The namespace to seach for (for example, 'name', 'cid',
        'smiles', ...). Default: 'name'

    Returns
    -------
    result : float
        The return value is the 'MolecularWeight' item of the json dict returned by PubChem or None
        if an error occurred.
    """
    # Insert 's' and 'namespace' into the URL.
    url = f"https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/{namespace}/{s}/property/MolecularWeight/json"
    # Get the json dict, return None if an error occurred.
    js = get_pubchem_json(url)
    if js is None:
        return None
    if len(js["PropertyTable"]["Properties"]) > 0:
        # Return the MolecularWeight property.
        return float(js["PropertyTable"]["Properties"][0]["MolecularWeight"])
    else:
        return None
        
def get_pubchem_properties(cids, props):
    """
    This method uses the PubChem REST API service to retrieve a table of
    compound properties for one or more compounds based on their cid(s).

    Parameters
    ----------
    cids : int, str or list
        The PubChem ID(s) of the compound(s) to look up.
    props : str or list
        The names of the properties to return. For possible values see
        https://pubchem.ncbi.nlm.nih.gov/docs/pug-rest#section=Compound-Property-Tables

    Returns
    -------
    result : list
        A list with the requested properties, or None if the request
        was unsuccessful.
    """
    # Ensure that cids is a list.
    if not isinstance(cids, list):
        cids = [cids]
    # Convert all elements of cids to strings.
    cids_str = [str(cid) for cid in cids]
    # Ensure that props is a list.
    if not isinstance(props, list):
        props = [props]
    # Check if each element in props is a value accepted by the API. Otherwise 
    # return None.
    for p in props:
        if not (p in PUBCHEM_COMPOUND_PROPS):
            print(f"Property {p} not allowed in PUBCHEM query")
            return None
    # Insert the cid(s) and propert(y/ies) into the URL.
    url = f"https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/cid/{','.join(cids_str)}/property/{','.join(props)}/json"
    # Get the json dict
    js = get_pubchem_json(url)
    # If the request was unsuccessful js will be None, so return None.
    if js is None:
        return None
    # If successful return the 'Properties' dict item.
    return js["PropertyTable"]["Properties"]
